from json_database.utils import fuzzy_match, match_one
from json_database import JsonDatabase, JsonStorageXDG


class Query:
    def __init__(self, db):
        if isinstance(db, JsonDatabase):
            self.result = list(db)
        else:
            self.result = [db]

    def contains_key(self, key, fuzzy=False, thresh=0.7):
        if fuzzy:
            after = []
            for e in self.result:
                filter = True
                for k in e:
                    score = fuzzy_match(k, key)
                    if score < thresh:
                        continue
                    filter = False
                if not filter:
                    after.append(e)
            self.result = after
        else:
            self.result = [a for a in self.result if a.get(key)]
        return self

    def contains_value(self, key, value, fuzzy=False, thresh=0.75):
        self.contains_key(key)
        if fuzzy:
            after = []
            for e in self.result:
                if isinstance(e[key], str):
                    score = fuzzy_match(value, e[key])
                    if score > thresh:
                        after.append(e)
                elif isinstance(e[key], list) or isinstance(e[key], dict):
                    v, score = match_one(value, e[key])
                    if score < thresh:
                        continue
                    after.append(e)
            self.result = after
        else:
            self.result = [a for a in self.result if value in a[key]]
        return self

    def value_contains(self, key, value, ignore_case=False):
        self.contains_key(key)
        if ignore_case:
            after = []
            value = str(value).lower()
            for e in self.result:
                if isinstance(e[key], str):
                    if value in e[key].lower():
                        after.append(e)
                elif isinstance(e[key], list):
                    if value in [str(_).lower() for _ in e[key]]:
                        after.append(e)
                elif isinstance(e[key], dict):
                    if value in [str(_).lower() for _ in e[key].keys()]:
                        after.append(e)
            self.result = after
        else:
            self.result = [e for e in self.result if value in e[key]]
        return self

    def value_contains_token(self, key, value, ignore_case=False):
        self.contains_key(key)
        after = []
        value = str(value)
        for e in self.result:
            if isinstance(e[key], str):
                if ignore_case:
                    if value.lower() in e[key].lower().split(" "):
                        after.append(e)
                else:
                    if value in e[key].split(" "):
                        after.append(e)
            elif value in e[key]:
                after.append(e)
        self.result = after
        return self

    def equal(self, key, value):
        self.contains_key(key)
        self.result = [a for a in self.result if a[key] == value]
        return self

    def bellow(self, key, value):
        self.contains_key(key)
        self.result = [a for a in self.result if a[key] < value]
        return self

    def above(self, key, value):
        self.contains_key(key)
        self.result = [a for a in self.result if a[key] > value]
        return self

    def bellow_or_equal(self, key, value):
        self.contains_key(key)
        self.result = [a for a in self.result if a[key] <= value]
        return self

    def above_or_equal(self, key, value):
        self.contains_key(key)
        self.result = [a for a in self.result if a[key] >= value]
        return self

    def in_range(self, key, min_value, max_value):
        self.contains_key(key)
        self.result = [a for a in self.result if min_value < a[key] < max_value]
        return self

    def all(self):
        return self

    def build(self):
        return self.result





import json
from difflib import SequenceMatcher


def fuzzy_match(x, against):
    """Perform a 'fuzzy' comparison between two strings.
    Returns:
        float: match percentage -- 1.0 for perfect match,
               down to 0.0 for no match at all.
    """
    return SequenceMatcher(None, x, against).ratio()


def match_one(query, choices):
    """
        Find best match from a list or dictionary given an input

        Arguments:
            query:   string to test
            choices: list or dictionary of choices

        Returns: tuple with best match, score
    """
    if isinstance(choices, dict):
        _choices = list(choices.keys())
    elif isinstance(choices, list):
        _choices = choices
    else:
        raise ValueError('a list or dict of choices must be provided')

    best = (_choices[0], fuzzy_match(query, _choices[0]))
    for c in _choices[1:]:
        score = fuzzy_match(query, c)
        if score > best[1]:
            best = (c, score)

    if isinstance(choices, dict):
        return (choices[best[0]], best[1])
    else:
        return best


def merge_dict(base, delta, merge_lists=True, skip_empty=True, no_dupes=True):
    """
        Recursively merging configuration dictionaries.

        Args:
            base:  Target for merge
            delta: Dictionary to merge into base
            merge_lists: if a list is found merge contents instead of replacing
            skip_empty: if an item in delta is empty, dont overwrite base
            no_dupes: when merging lists deduplicate entries
    """

    for k, d in delta.items():
        b = base.get(k)
        if isinstance(d, dict) and isinstance(b, dict):
            merge_dict(b, d)
        else:
            if skip_empty and not d:
                # dont replace if new entry is empty
                pass
            elif all((isinstance(b, list), isinstance(d, list), merge_lists)):
                base[k] += d
                if no_dupes:
                    # deduplicate
                    base[k] = list(set(base[k]))
            else:
                base[k] = d
    return base


def load_commented_json(filename):
    """ Loads an JSON file, ignoring comments

    Supports a trivial extension to the JSON file format.  Allow comments
    to be embedded within the JSON, requiring that a comment be on an
    independent line starting with '//' or '#'.

    NOTE: A file created with these style comments will break strict JSON
          parsers.  This is similar to but lighter-weight than "human json"
          proposed at https://hjson.org

    Args:
        filename (str):  path to the commented JSON file

    Returns:
        obj: decoded Python object
    """
    with open(filename) as f:
        contents = f.read()

    return json.loads(uncomment_json(contents))


def uncomment_json(commented_json_str):
    """ Removes comments from a JSON string.

    Supporting a trivial extension to the JSON format.  Allow comments
    to be embedded within the JSON, requiring that a comment be on an
    independent line starting with '//' or '#'.

    Example...
       {
         // comment
         'name' : 'value'
       }

    Args:
        commented_json_str (str):  a JSON string

    Returns:
        str: uncommented, legal JSON
    """
    lines = commented_json_str.splitlines()
    # remove all comment lines, starting with // or #
    nocomment = []
    for line in lines:
        stripped = line.lstrip()
        if stripped.startswith("//") or stripped.startswith("#"):
            continue
        nocomment.append(line)

    return " ".join(nocomment)


def is_jsonifiable(thing):
    if not isinstance(thing, dict):
        if isinstance(thing, str):
            try:
                json.loads(thing)
                return True
            except:
                pass
        else:
            try:
                thing.__dict__
                return True
            except:
                pass
        return False
    return True


def get_key_recursively(search_dict, field, filter_None=True):
    """
    Takes a dict with nested lists and dicts,
    and searches all dicts for a key of the field
    provided.
    """
    if not is_jsonifiable(search_dict):
        raise ValueError("unparseable format")
    fields_found = []

    for key, value in search_dict.items():
        if value is None and filter_None:
            continue
        if key == field:
            fields_found.append(search_dict)

        elif isinstance(value, dict):
            fields_found += get_key_recursively(value, field, filter_None)

        elif isinstance(value, list):
            for item in value:
                if not isinstance(item, dict):
                    try:
                        if get_key_recursively(item.__dict__, field, filter_None):
                            fields_found.append(item)
                    except:
                        continue  # can't parse
                else:
                    fields_found += get_key_recursively(item, field, filter_None)

    return fields_found


def get_key_recursively_fuzzy(search_dict, field, thresh=0.6, filter_None=True):
    """
    Takes a dict with nested lists and dicts,
    and searches all dicts for a key of the field
    provided.
    """
    if not is_jsonifiable(search_dict):
        raise ValueError("unparseable format")

    fields_found = []

    for key, value in search_dict.items():
        if value is None and filter_None:
            continue
        score = 0
        if isinstance(key, str):
            score = fuzzy_match(key, field)

        if score >= thresh:
            fields_found.append((search_dict, score))
        elif isinstance(value, dict):
            fields_found += get_key_recursively_fuzzy(value, field, thresh, filter_None)

        elif isinstance(value, list):
            for item in value:
                if not isinstance(item, dict):
                    try:
                        if get_key_recursively_fuzzy(item.__dict__, field, thresh, filter_None):
                            fields_found.append((item, score))
                    except:
                        continue  # can't parse
                else:
                    fields_found += get_key_recursively_fuzzy(item, field, thresh, filter_None)
    return sorted(fields_found, key = lambda i: i[1],reverse=True)


def get_value_recursively(search_dict, field, target_value):
    """
    Takes a dict with nested lists and dicts,
    and searches all dicts for a key of the field
    provided.
    """
    if not is_jsonifiable(search_dict):
        raise ValueError("unparseable format")
    fields_found = []

    for key, value in search_dict.items():

        if key == field and value == target_value:
            fields_found.append(search_dict)

        elif isinstance(value, dict):
            fields_found += get_value_recursively(value, field, target_value)

        elif isinstance(value, list):
            for item in value:
                if not isinstance(item, dict):
                    try:
                        if get_value_recursively(item.__dict__, field, target_value):
                            fields_found.append(item)
                    except:
                        continue  # can't parse
                else:
                    fields_found += get_value_recursively(item, field, target_value)

    return fields_found


def get_value_recursively_fuzzy(search_dict, field, target_value, thresh=0.6):
    """
    Takes a dict with nested lists and dicts,
    and searches all dicts for a key of the field
    provided.
    """
    if not is_jsonifiable(search_dict):
        raise ValueError("unparseable format")
    fields_found = []
    for key, value in search_dict.items():
        if key == field:
            if isinstance(value, str):
                score = fuzzy_match(target_value, value)
                if score >= thresh:
                    fields_found.append((search_dict, score))
            elif isinstance(value, list):
                for item in value:
                    score = fuzzy_match(target_value, item)
                    if score >= thresh:
                        fields_found.append((search_dict, score))
        elif isinstance(value, dict):
            fields_found += get_value_recursively_fuzzy(value, field, target_value, thresh)

        elif isinstance(value, list):
            for item in value:
                if not isinstance(item, dict):
                    try:
                        found = get_value_recursively_fuzzy(item.__dict__, field, target_value, thresh)
                        if len(found):
                            fields_found.append((item, found[0][1]))
                    except:
                        continue  # can't parse
                else:
                    fields_found += get_value_recursively_fuzzy(item, field, target_value, thresh)

    return sorted(fields_found, key = lambda i: i[1],reverse=True)


def jsonify_recursively(thing):
    if isinstance(thing, list):
        jsonified = list(thing)
        for idx, item in enumerate(thing):
            jsonified[idx] = jsonify_recursively(item)
    elif isinstance(thing, dict):
        try:
            # can't import at top level to do proper check
            jsonified = dict(thing.db)
        except:
            jsonified = dict(thing)
        for key in jsonified.keys():
            value = jsonified[key]
            jsonified[key] = jsonify_recursively(value)
    else:
        try:
            jsonified = thing.__dict__
        except:
            jsonified = thing
    return jsonified



class InvalidItemID(ValueError):
    """ ItemID is invalid """


class DatabaseNotCommitted(FileNotFoundError):
    """ Database has not been saved in yet """


class SessionError(RuntimeError):
    """ Could not commit database"""


class MatchError(ValueError):
    """ could not match an item in db """
